import os
import sys
import types
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union, cast

import click
from click.exceptions import Exit
from click.utils import _expand_args
from pyappconf import AppConfig, BaseConfig
from typer import Typer
from typer.main import get_command as typer_get_command
from typing_extensions import TypeGuard

from cliconf.arg_store import ARGS_STORE
from cliconf.command_name import get_command_name
from cliconf.logger import log
from cliconf.options import create_generate_config_option
from cliconf.settings import CLIConfSettings


class CLIConfCallable:
    pyappconf_settings: AppConfig
    cliconf_settings: CLIConfSettings
    model_cls: Type[BaseConfig]

    def __call__(self, *args, **kwargs):
        ...


def get_command(typer_instance: Typer) -> click.Command:
    """
    Extends typer's get_command function to modify the created click.Command instance
    to inspect the passed arguments and load from config.
    """
    command = cast(Union[click.Command, click.Group], typer_get_command(typer_instance))

    if _is_command_group(command):
        for subcommand in command.commands.values():
            _customize_command(subcommand)
        # Override the main function to load config
        command.main = types.MethodType(_cli_conf_main_multi_command, command)
        return command

    # Single command
    _customize_command(command)
    # Override the main function to load config
    command.main = types.MethodType(_cli_conf_main_single_command, command)

    return command


def _is_command_group(
    command: Union[click.Command, click.Group]
) -> TypeGuard[click.Group]:
    return hasattr(command, "commands")


def _is_cliconf_command(command: click.Command) -> bool:
    return hasattr(command.callback, "cliconf_settings")


def _is_cliconf_callable(fn: Callable) -> TypeGuard[CLIConfCallable]:
    return (
        hasattr(fn, "cliconf_settings")
        and hasattr(fn, "pyappconf_settings")
        and hasattr(fn, "model_cls")
        and callable(fn)
    )


def _customize_command(
    command: click.Command,
):
    if not _is_cliconf_command(command):
        log.debug(f"Not a cliconf command, skipping customization: {command.name}")
        return

    callback = command.callback

    if not _is_cliconf_callable(callback):
        return

    pyappconf_settings: AppConfig = callback.pyappconf_settings
    cliconf_settings: CLIConfSettings = callback.cliconf_settings
    model_cls: Type[BaseConfig] = callback.model_cls
    command.params.append(
        create_generate_config_option(
            pyappconf_settings.supported_formats,
            pyappconf_settings.default_format,
            model_cls,
            callback,  # type: ignore
            cliconf_settings.generate_config_option_name,
        )
    )


def _cli_conf_main_single_command(
    self: click.Command,
    args: Optional[Sequence[str]] = None,
    prog_name: Optional[str] = None,
    complete_var: Optional[str] = None,
    standalone_mode: bool = True,
    windows_expand_args: bool = True,
    **extra: Any,
):
    """
    A modified version of click.Command's main function that records which arguments were passed
    """
    use_args = _get_arguments_from_passed_or_argv(args)
    func_name = prog_name or get_command_name(self.callback.__name__)  # type: ignore
    params = _create_passed_param_dict_from_command(self, func_name, use_args)
    # It seems typer always provides prog_name, but for safety calculate a fallback
    ARGS_STORE.add_command(func_name, use_args, params)
    return super(type(self), self).main(  # type: ignore
        args, func_name, complete_var, standalone_mode, windows_expand_args, **extra
    )


def _cli_conf_main_multi_command(
    self: click.Group,
    args: Optional[Sequence[str]] = None,
    prog_name: Optional[str] = None,
    complete_var: Optional[str] = None,
    standalone_mode: bool = True,
    windows_expand_args: bool = True,
    **extra: Any,
):
    """
    A modified version of click.Group's main function that records which arguments were passed
    """

    def run_original():
        return super(type(self), self).main(  # type: ignore
            args, prog_name, complete_var, standalone_mode, windows_expand_args, **extra
        )

    use_args = _get_arguments_from_passed_or_argv(args)
    if len(use_args) == 0:
        # No arguments passed. No need to configure arguments
        return run_original()
    sub_command_name, sub_command_args = use_args[0], use_args[1:]
    sub_command = self.commands.get(sub_command_name)
    if sub_command is None:
        # First argument did not match a subcommand. Must be requesting help, no need to configure arguments
        return run_original()
    func_name = get_command_name(sub_command.callback.__name__)  # type: ignore
    params = _create_passed_param_dict_from_command(
        sub_command, func_name, sub_command_args
    )
    ARGS_STORE.add_command(func_name, sub_command_args, params)
    return super(type(self), self).main(  # type: ignore
        args, func_name, complete_var, standalone_mode, windows_expand_args, **extra
    )


def _create_passed_param_dict_from_command(
    command: click.Command,
    prog_name: str,
    args: Sequence[str],
) -> Dict[str, Any]:
    # Click's Command.make_context will raise Exit if the arguments are invalid
    # Now we are parsing the arguments before click does, so we must handle the Exit.
    # Here we just exit with the exit code, just as Click does in standalone_mode
    try:
        context = command.make_context(prog_name, [*args])
    except Exit as e:
        sys.exit(e.exit_code)
    parser = command.make_parser(context)
    opts, _, param_order = parser.parse_args(args=[*args])
    # Reorder the opts dict to match the order of the command's params
    out_opts: Dict[str, Any] = {}
    for argument in param_order:
        if argument.name in opts:
            out_opts[argument.name] = opts[argument.name]
    return out_opts


def _get_arguments_from_passed_or_argv(
    args: Optional[Sequence[str]] = None,
) -> List[str]:
    """
    Returns the arguments passed to the command.

    Note: Mostly adapted from click.BaseCommand.main
    :param args:
    :return:
    """
    if args is not None:
        return list(args)

    args = sys.argv[1:]

    if os.name == "nt":
        # It's not ideal to be using a private method, but want to make sure
        # it works exactly the same for param extraction as how Click handles it
        return _expand_args(args)
    return args



import warnings
from types import FunctionType
from typing import Any, Dict, Optional, Sequence, Type, no_type_check

from pyappconf import AppConfig, BaseConfig
from pydantic import create_model

from cliconf.arg_store import ARGS_STORE
from cliconf.command_name import get_command_name
from cliconf.dynamic_config_name import dynamic_model_class_name
from cliconf.ext_inspect import get_function_params
from cliconf.ext_pyappconf import create_cli_base_config_class


@no_type_check
def create_dynamic_config_class_from_function(
    func: FunctionType,
    settings: AppConfig,
    base_cls: Optional[Type[BaseConfig]] = None,
    model_is_injected: bool = False,
    make_optional: bool = True,
) -> Type[BaseConfig]:
    """
    Create a BaseConfig class from a function.
    """
    base_cls = base_cls or create_cli_base_config_class(BaseConfig, settings)
    params = get_function_params(
        func, make_optional=make_optional, model_is_injected=model_is_injected
    )

    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore",
            category=RuntimeWarning,
            message='fields may not start with an underscore, ignoring "_settings"',
        )

        model_cls = create_model(
            dynamic_model_class_name(func),
            __base__=base_cls,
            **params,
            settings=settings,
            _settings=settings,
        )
    return model_cls


def filter_func_args_and_kwargs_to_get_user_passed_data(
    func: FunctionType,
    func_args: Sequence[Any],
    func_kwargs: Dict[str, Any],
) -> Dict[str, Any]:
    args_kwargs = dict(zip(func.__code__.co_varnames[1:], func_args))
    args_kwargs.update(func_kwargs)
    # Get user passed args from command line via args store
    args_store = ARGS_STORE[get_command_name(func.__name__)]
    user_kwargs = args_store.params
    return user_kwargs



from typing import Optional, Type

from pyappconf import BaseConfig
from pydantic import BaseModel


class CLIConfSettings(BaseModel):
    """
    Settings for the CLI configuration that are specific to cliconf
    """

    generate_config_option_name: str = "config-gen"
    base_cls: Optional[Type[BaseConfig]] = None
    make_fields_optional: bool = True
    recursive_loading: bool = False
    inject_model: bool = False


DEFAULT_SETTINGS = CLIConfSettings()



import functools
import inspect
from types import FunctionType
from typing import Any, Callable, Dict, List, Type

from pyappconf import AppConfig, BaseConfig

from cliconf.arg_store import ARGS_STORE
from cliconf.command_name import get_command_name
from cliconf.dynamic_config import (
    create_dynamic_config_class_from_function,
    filter_func_args_and_kwargs_to_get_user_passed_data,
)
from cliconf.ext_typer import get_arg_names_that_can_be_processed_by_typer
from cliconf.py_api import execute_cliconf_func_as_python_func
from cliconf.settings import DEFAULT_SETTINGS, CLIConfSettings


class _ModelContainer:
    def __init__(self):
        # Use a mutable container so that function can mutate
        self._models: List[Type[BaseConfig]] = []

    @property
    def model_cls(self) -> Type[BaseConfig]:
        return self._models[0]

    def set_model_cls(self, model_cls: Type[BaseConfig]):
        if len(self._models) > 0:
            self._models.pop()
        self._models.append(model_cls)

    model: Type[BaseConfig]


def _is_executing_from_cli(func: FunctionType) -> bool:
    return ARGS_STORE.args_are_stored_for(get_command_name(func.__name__))


def configure(
    pyappconf_settings: AppConfig,
    cliconf_settings: CLIConfSettings = DEFAULT_SETTINGS,
) -> Callable:
    def actual_decorator(func: FunctionType):
        model_cls = create_dynamic_config_class_from_function(
            func,
            pyappconf_settings,
            cliconf_settings.base_cls,
            model_is_injected=cliconf_settings.inject_model,
            make_optional=cliconf_settings.make_fields_optional,
        )

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not _is_executing_from_cli(func):
                return execute_cliconf_func_as_python_func(
                    func, cliconf_settings.inject_model, *args, **kwargs
                )

            # Load the config, overriding with any user passed args
            user_passed_data = filter_func_args_and_kwargs_to_get_user_passed_data(
                func, args, kwargs
            )
            if cliconf_settings.recursive_loading:
                try:
                    config = model_cls.load_recursive(model_kwargs=user_passed_data)
                except FileNotFoundError:
                    config = model_cls.load_or_create(model_kwargs=user_passed_data)
            else:
                config = model_cls.load_or_create(model_kwargs=user_passed_data)
            # Clear the args store. This shouldn't matter in normal usage, but it helps
            # with testing.
            ARGS_STORE.remove_command(get_command_name(func.__name__))

            # If injecting model, provide that as the first argument
            if cliconf_settings.inject_model:
                return func(config, **model_as_dict(config))

            return func(**model_as_dict(config))

        # Attach the generated config model class to the function, so it can be imported in
        # the py config format
        wrapper.model_cls = model_cls  # type: ignore

        # Also attach the settings to the function, so it can be used by the typer instance
        # to customize the options to add the --config-gen option
        wrapper.pyappconf_settings = pyappconf_settings  # type: ignore
        wrapper.cliconf_settings = cliconf_settings  # type: ignore

        # Override call signature to exclude any variables that cannot be processed by typer
        # Otherwise typer will fail while trying to create the click command.
        # These excluded values will come only from pyappconf and not from CLI.
        typer_args = get_arg_names_that_can_be_processed_by_typer(func)
        sig = inspect.signature(func)
        typer_sig = sig.replace(
            parameters=tuple(
                val for name, val in sig.parameters.items() if name in typer_args
            )
        )
        wrapper.__signature__ = typer_sig  # type: ignore

        return wrapper

    return actual_decorator


def model_as_dict(model: BaseConfig) -> Dict[str, Any]:
    """
    Only serialize the top-layer of the model to a dict, leave nested models as Pydantic models
    """
    data: Dict[str, Any] = {}
    for name, field in model.__fields__.items():
        if name == "settings":
            continue
        data[name] = getattr(model, name)
    return data



from typing import Callable, Optional

import pytest
import typer
from pydantic import ValidationError

from cliconf import CLIConf, CLIConfSettings, configure
from tests.fixtures.app_settings import SETTINGS_TWO_PY
from tests.fixtures.cliconfs import default_func_for_single_command_py


def test_configure_creates_dynamic_model():
    cliconf_settings = CLIConfSettings(make_fields_optional=False)

    @configure(SETTINGS_TWO_PY, cliconf_settings=cliconf_settings)
    def my_cli_func(
        a: int,
        b: str = "b",
        c: Optional[str] = None,
        *,
        d: float,
        e: Optional[int],
        f: str = "f",
        g: Optional[str] = None
    ):
        pass

    model_cls = my_cli_func.model_cls
    model = model_cls(a=10, d=30, e=40)
    assert model.a == 10
    assert model.b == "b"
    assert model.c is None
    assert model.d == 30
    assert model.e == 40
    assert model.f == "f"
    assert model.g is None

    # Fields are required, so shouldn't be able to construct without args
    with pytest.raises(ValidationError):
        model_cls()


def test_configure_creates_dynamic_model_with_optional_fields():
    @configure(SETTINGS_TWO_PY)
    def my_cli_func(
        a: int,
        b: str = "b",
        c: Optional[str] = None,
        *,
        d: float,
        e: Optional[int],
        f: str = "f",
        g: Optional[str] = None
    ):
        pass

    model_cls = my_cli_func.model_cls
    model = model_cls(a=10, d=30, e=40)
    assert model.a == 10
    assert model.b == "b"
    assert model.c is None
    assert model.d == 30
    assert model.e == 40
    assert model.f == "f"
    assert model.g is None

    # Fields are not requires, so should be able to construct without args
    empty = model_cls()
    assert empty.a is None
    assert empty.b == "b"
    assert empty.c is None
    assert empty.d is None
    assert empty.e is None
    assert empty.f == "f"
    assert empty.g is None


def test_configure_creates_dynamic_model_with_typer():
    cliconf_instance = CLIConf(name="dynamic_model_with_typer")

    @cliconf_instance.command()
    @configure(pyappconf_settings=SETTINGS_TWO_PY)
    def my_cli_func(
        a: str,
        b: int = typer.Argument(..., help="b help"),
        c: float = typer.Option(3.2, help="c help"),
        d: Callable[[float], None] = default_func_for_single_command_py,
    ):
        print(a, b, c, d(c))

    model_cls = my_cli_func.model_cls
    model = model_cls(a="a", b=1000)
    assert model.a == "a"
    assert model.b == 1000
    assert model.c == 3.2
    assert model.d == default_func_for_single_command_py



from cliconf.configure import model_as_dict
from tests.fixtures.cliconfs import (
    my_cli_func_one_injected_model_yaml,
    my_cli_func_one_yaml,
)


def test_cliconf_decorated_executes_as_a_normal_python_function():
    result = my_cli_func_one_yaml("a", 2)
    assert result == ("a", 2, 3.2)


def test_cliconf_decorated_with_inject_model_executes_as_a_normal_python_function():
    mod = my_cli_func_one_injected_model_yaml.model_cls(a="a", b=2)
    result = my_cli_func_one_injected_model_yaml(mod, **model_as_dict(mod))
    assert result == mod



from pathlib import Path

TESTS_DIR = Path(__file__).parent
INPUT_FILES_DIR = TESTS_DIR / "input_files"
CONFIGS_DIR = INPUT_FILES_DIR / "configs"
PLAIN_CONFIGS_DIR = CONFIGS_DIR / "plain"
OVERRIDES_CONFIGS_DIR = CONFIGS_DIR / "overrides"
NESTED_OVERRIDES_CONFIGS_DIR = OVERRIDES_CONFIGS_DIR / "nested"

CONFIG_ONE = CONFIGS_DIR / "one.yaml"



from types import FunctionType
from typing import Any, Dict, Sequence, Type

import click
from click import Context, Parameter
from pyappconf import BaseConfig, ConfigFormats

from cliconf.ext_pyappconf import save_model
from cliconf.settings import DEFAULT_SETTINGS


def create_generate_config_option(
    supported_formats: Sequence[ConfigFormats],
    default_format: ConfigFormats,
    model_cls: Type[BaseConfig],
    func: FunctionType,
    option_name: str = DEFAULT_SETTINGS.generate_config_option_name,
) -> click.Option:
    def gen_config(ctx: Context, param: Parameter, value: str) -> None:
        if len(supported_formats) == 1:
            use_format = default_format
        else:
            use_format = value

        if value and not ctx.resilient_parsing:
            if use_format != default_format:
                new_settings = model_cls._settings.copy(default_format=use_format)
                model_obj = model_cls(settings=new_settings)
            else:
                model_obj = model_cls()
            click.echo(f"Saving config to {model_obj.settings.config_location}")
            save_model(model_obj, func)
            ctx.exit()

    opt_flag = [f"--{option_name}"]
    common_kwargs: Dict[str, Any] = dict(
        help="Generate a config file from the command line arguments.",
        callback=gen_config,
    )

    if len(supported_formats) == 1:
        return click.Option(
            opt_flag,
            is_flag=True,
            **common_kwargs,
        )

    return click.Option(
        opt_flag,
        type=click.Choice(supported_formats),
        prompt="What format would you like to generate the config file in?",
        prompt_required=False,
        **common_kwargs,
    )



from types import FunctionType
from typing import Any

from cliconf.ext_inspect import get_function_args, get_function_params


def execute_cliconf_func_as_python_func(
    func: FunctionType, model_is_injected: bool, *args, **kwargs
) -> Any:
    """
    Execute a cliconf function as a normal python function.

    Unwraps any typer options and arguments that may be used as defaults.
    """
    arg_spec = get_function_args(func)
    user_passed_arg_names = [arg_name for arg, arg_name in zip(args, arg_spec)]
    params = get_function_params(func, model_is_injected=model_is_injected)
    defaults_from_func = {name: value for name, (_, value) in params.items()}
    defaults_without_args = {
        key: value
        for key, value in defaults_from_func.items()
        if key not in user_passed_arg_names
    }
    new_kwargs = {**defaults_without_args, **kwargs}
    return func(*args, **new_kwargs)



from pathlib import Path
from typing import Any, Callable, Optional, Tuple

import pytest
import typer
from pydantic import BaseModel

from cliconf import CLIConfSettings, configure
from cliconf.main import CLIConf
from tests.dirutils import create_temp_path
from tests.fixtures.app_settings import (
    SETTINGS_ALL_OPTIONAL_JSON,
    SETTINGS_NESTED_CONFIG_YAML,
    SETTINGS_ONE_MULTI_FORMAT,
    SETTINGS_ONE_RECURSIVE_YAML,
    SETTINGS_ONE_YAML,
    SETTINGS_TWO_PY,
)

single_command_yaml_cliconf = CLIConf(name="single_command_yaml")


@single_command_yaml_cliconf.command()
@configure(pyappconf_settings=SETTINGS_ONE_YAML)
def my_cli_func_one_yaml(
    a: str,
    b: int = typer.Argument(..., help="b help"),
    c: float = typer.Option(3.2, help="c help"),
):
    print(a, b, c)


single_command_recursive_yaml_cliconf = CLIConf(name="single_command_recursive_yaml")


@single_command_recursive_yaml_cliconf.command()
@configure(
    pyappconf_settings=SETTINGS_ONE_RECURSIVE_YAML,
    cliconf_settings=CLIConfSettings(recursive_loading=True),
)
def my_cli_func_one_recursive_yaml(
    a: str,
    b: int = typer.Argument(..., help="b help"),
    c: float = typer.Option(3.2, help="c help"),
):
    print(a, b, c)


@pytest.fixture
def single_command_yaml_cliconf_in_temp_dir() -> Tuple[CLIConf, Path]:
    with create_temp_path() as temp_path:
        settings = SETTINGS_ONE_YAML.copy(custom_config_folder=temp_path)
        temp_dir_cliconf = CLIConf(name="single_command_yaml_in_temp_dir")

        @temp_dir_cliconf.command()
        @configure(pyappconf_settings=settings)
        def my_cli_func_one_yaml(
            a: str,
            b: int = typer.Argument(..., help="b help"),
            c: float = typer.Option(3.2, help="c help"),
        ):
            print(a, b, c)

        yield temp_dir_cliconf, temp_path


single_command_multi_format_cliconf = CLIConf(name="single_command_multi_format")


@single_command_multi_format_cliconf.command()
@configure(pyappconf_settings=SETTINGS_ONE_MULTI_FORMAT)
def my_cli_func_one_yaml(
    a: str,
    b: int = typer.Argument(..., help="b help"),
    c: float = typer.Option(3.2, help="c help"),
):
    print(a, b, c)
    return a, b, c


single_command_nested_config_yaml_cliconf = CLIConf(
    name="single_command_nested_config_yaml"
)


class NestedConfig(BaseModel):
    na: str
    nb: float


@single_command_nested_config_yaml_cliconf.command()
@configure(pyappconf_settings=SETTINGS_NESTED_CONFIG_YAML)
def my_cli_func_one_nested_config_yaml(
    a: str,
    b: Optional[NestedConfig] = None,
):
    print(a, b)


multi_command_shared_config_yaml_cliconf = CLIConf(name="multi_command_yaml")


@multi_command_shared_config_yaml_cliconf.command(name="one", help="one help")
@configure(pyappconf_settings=SETTINGS_ONE_YAML)
def my_cli_func_one_multi_yaml(
    a: str,
    b: int = typer.Argument(..., help="b help"),
    c: float = typer.Option(3.2, help="c help"),
):
    print(a, b, c)


@multi_command_shared_config_yaml_cliconf.command(name="two", help="two help")
@configure(pyappconf_settings=SETTINGS_ONE_YAML)
def my_cli_func_two_multi_yaml(
    a: str,
    b: int = typer.Argument(..., help="b help"),
    d: float = typer.Option(13.2, help="d help"),
):
    print(d, b, a)


single_command_py_cliconf = CLIConf(name="single_command_py")


def default_func_for_single_command_py(c: float) -> str:
    return f"default {c}"


@single_command_py_cliconf.command()
@configure(pyappconf_settings=SETTINGS_TWO_PY)
def my_cli_func_two_py(
    a: str,
    b: int = typer.Argument(..., help="b help"),
    c: float = typer.Option(3.2, help="c help"),
    d: Callable[[float], str] = default_func_for_single_command_py,
):
    print(a, b, c, d(c))


@pytest.fixture
def single_command_py_cliconf_in_temp_dir() -> Tuple[CLIConf, Path]:
    with create_temp_path() as temp_path:
        settings = SETTINGS_TWO_PY.copy(custom_config_folder=temp_path)
        temp_dir_cliconf = CLIConf(name="single_command_py_in_temp_dir")

        @temp_dir_cliconf.command()
        @configure(pyappconf_settings=settings)
        def my_cli_func_two_py(
            a: str,
            b: int = typer.Argument(..., help="b help"),
            c: float = typer.Option(3.2, help="c help"),
            d: Callable[[float], str] = default_func_for_single_command_py,
        ):
            print(a, b, c, d(c))

        yield temp_dir_cliconf, temp_path


single_command_all_optional_json_cliconf = CLIConf(
    name="single_command_all_optional_json"
)


@single_command_all_optional_json_cliconf.command()
@configure(pyappconf_settings=SETTINGS_ALL_OPTIONAL_JSON)
def my_cli_func_all_optional_json(
    a: str = typer.Option("abc", help="a help"),
    b: int = typer.Option(123, help="b help"),
):
    print(a, b)


single_command_injected_model_yaml_cliconf = CLIConf(
    name="single_command_injected_model_yaml"
)


@single_command_injected_model_yaml_cliconf.command()
@configure(
    pyappconf_settings=SETTINGS_ONE_YAML,
    cliconf_settings=CLIConfSettings(inject_model=True),
)
def my_cli_func_one_injected_model_yaml(
    model,
    a: str,
    b: int = typer.Argument(..., help="b help"),
    c: float = typer.Option(3.2, help="c help"),
):
    print(model, a, b, c)
    return model


if __name__ == "__main__":
    single_command_py_cliconf()



from types import FunctionType


def dynamic_model_class_name(func: FunctionType) -> str:
    # Convert the name of the function from snake_case to PascalCase
    func_in_class_style_name = "".join(
        part.title() for part in func.__name__.split("_")
    )
    return f"{func_in_class_style_name}Config"



import re
import shlex
from pathlib import Path
from typing import Sequence, Tuple

from click.testing import Result

from cliconf.main import CLIConf
from cliconf.testing import CLIRunner
from tests import ext_click
from tests.config import CONFIGS_DIR, NESTED_OVERRIDES_CONFIGS_DIR, PLAIN_CONFIGS_DIR
from tests.dirutils import change_directory_to
from tests.fixtures.cliconfs import (
    multi_command_shared_config_yaml_cliconf,
    single_command_injected_model_yaml_cliconf,
    single_command_multi_format_cliconf,
    single_command_nested_config_yaml_cliconf,
    single_command_py_cliconf,
    single_command_py_cliconf_in_temp_dir,
    single_command_recursive_yaml_cliconf,
    single_command_yaml_cliconf,
    single_command_yaml_cliconf_in_temp_dir,
)

runner = CLIRunner()

ansi_escape = re.compile(
    r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
""",
    re.VERBOSE,
)


def strip_all_ansi(string: str) -> str:
    return ansi_escape.sub("", string)


class CLIRunnerException(Exception):
    pass


def run(instance: CLIConf, args: Sequence[str]) -> Result:
    result = runner.invoke(instance, args)
    if result.exit_code != 0:
        output = ext_click.result_to_message(result)
        command = shlex.join([instance.info.name, *args])
        raise CLIRunnerException(
            f"{command} with exited with code {result.exit_code}.\n{output}"
        )
    return result


def test_single_command_cliconf_reads_from_yaml_config():
    result = run(single_command_yaml_cliconf, ["a", "2"])
    assert result.stdout == "a 2 45.6\n"


def test_recursive_single_command_cliconf_reads_recursively_from_yaml_config():
    with change_directory_to(NESTED_OVERRIDES_CONFIGS_DIR):
        result = run(single_command_recursive_yaml_cliconf, ["a", "2"])
        assert result.stdout == "a 2 45.6\n"


def test_recursive_single_command_cliconf_loads_default_config_when_none_found():
    with change_directory_to(CONFIGS_DIR):
        result = run(single_command_recursive_yaml_cliconf, ["a", "2"])
        assert result.stdout == "a 2 3.2\n"


def test_single_command_multi_format_cliconf_reads_from_yaml_config():
    result = run(single_command_multi_format_cliconf, ["a", "2"])
    assert result.stdout == "a 2 45.6\n"


def test_single_command_cliconf_reads_nested_config_from_yaml_config():
    result = run(single_command_nested_config_yaml_cliconf, ["a"])
    assert result.stdout == "a na='abc' nb=123.0\n"


def test_multi_command_shared_config_cliconf_reads_from_yaml_config():
    result_one = run(multi_command_shared_config_yaml_cliconf, ["one", "a", "2"])
    result_two = run(multi_command_shared_config_yaml_cliconf, ["two", "a", "2"])
    assert result_one.stdout == "a 2 45.6\n"
    assert result_two.stdout == "13.2 2 a\n"


def _has_line_containing_each(text: str, *segments: Sequence[str]) -> bool:
    lines = strip_all_ansi(text).splitlines()
    for line in lines:
        match = True
        for segment in segments:
            if segment not in line:
                match = False
                continue

        if match:
            # Must have all segments in text
            return True
    return False


def test_single_command_cliconf_prints_help():

    result = run(single_command_yaml_cliconf, ["--help"])

    def has(*segments: Sequence[str]) -> bool:
        found = _has_line_containing_each(result.stdout, *segments)
        if not found:
            print(segments, "not found in")
            print(result.stdout)
        return found

    assert has("a", "TEXT", "[default: None]", "[required]")
    assert has("b", "INTEGER", "b help", "[default: None]", "[required]")
    assert has("--c", "FLOAT")
    assert has("c help", "[default: 3.2]")
    assert has("--help")


def test_multi_command_cliconf_prints_help():

    result = run(multi_command_shared_config_yaml_cliconf, ["--help"])

    def has(*segments: Sequence[str]) -> bool:
        found = _has_line_containing_each(result.stdout, *segments)
        if not found:
            print(segments, "not found in")
            print(result.stdout)
        return found

    assert has("one", "one help")
    assert has("two", "two help")


def test_single_command_cliconf_reads_py_config():
    result = run(single_command_py_cliconf, ["a", "2"])
    assert result.stdout == "a 2 123.4 custom 123.4\n"


def test_single_command_cliconf_reads_from_environment_over_config(monkeypatch):
    monkeypatch.setenv("MYAPP_C", "98.3")
    result = run(single_command_yaml_cliconf, ["a", "2"])
    assert result.stdout == "a 2 98.3\n"


def test_single_command_cliconf_writes_config_file_py(
    single_command_py_cliconf_in_temp_dir: Tuple[CLIConf, Path]
):
    cliconf_obj, temp_path = single_command_py_cliconf_in_temp_dir
    result = run(cliconf_obj, ["--config-gen"])
    assert "Saving config to" in result.stdout

    # Check that generated files are the same as in input_files
    for file_name in ("two.py", "two.pyi"):
        input_file = PLAIN_CONFIGS_DIR / file_name
        output_file = temp_path / file_name
        assert input_file.read_text() == output_file.read_text()


def test_single_command_cliconf_writes_config_file_yaml(
    single_command_yaml_cliconf_in_temp_dir: Tuple[CLIConf, Path]
):
    cliconf_obj, temp_path = single_command_yaml_cliconf_in_temp_dir
    result = run(cliconf_obj, ["--config-gen"])
    assert "Saving config to" in result.stdout

    # Check that generated files are the same as in input_files
    file_name = "one.yaml"
    input_file = PLAIN_CONFIGS_DIR / file_name
    output_file = temp_path / file_name
    assert input_file.read_text() == output_file.read_text()


def test_cliconf_injects_model():
    result = run(single_command_injected_model_yaml_cliconf, ["a", "2"])
    assert "a 2 45.6" in result.stdout
    assert "a='a' b=2 c=45.6" in result.stdout



import inspect
from types import FunctionType
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
    no_type_check,
)

from typing_extensions import TypeGuard

from cliconf.ext_typer import is_typer_parameter_info


@no_type_check
def get_function_params(
    func: FunctionType,
    model_is_injected: bool = False,
    make_optional: bool = True,
) -> Dict[str, Tuple[Type, Any]]:
    (
        args,
        _,
        varkw,
        defaults,
        kwonlyargs,
        kwonlydefaults,
        annotations,
    ) = inspect.getfullargspec(func)
    defaults = defaults or []
    args = args or []

    if model_is_injected:
        # Model is first argument, ensure it isn't recursively included in itself
        args = args[1:]

    non_default_args = len(args) - len(defaults)
    defaults = (...,) * non_default_args + defaults

    keyword_only_params = {
        param: (
            _extract_type(annotations.get(param), make_optional),
            _extract_default(kwonlydefaults.get(param), make_optional),
        )
        for param in kwonlyargs
    }
    params = {
        param: (
            _extract_type(annotations.get(param), make_optional),
            _extract_default(default, make_optional),
        )
        for param, default in zip(args, defaults)
    }
    return {
        **params,
        **keyword_only_params,
    }


def get_function_args(func: FunctionType) -> List[str]:
    (
        args,
        _,
        _,
        _,
        _,
        _,
        _,
    ) = inspect.getfullargspec(func)
    return args or []


@no_type_check
def _extract_type(typ: Optional[type], make_optional: bool) -> type:
    if typ is None:
        return Any

    if not make_optional:
        return typ

    if _is_optional(typ):
        return typ

    # Wrap type in Optional if it is not already
    return Optional[typ]


def _extract_from_typer_parameter_info_if_necessary(value: Any) -> Any:
    if is_typer_parameter_info(value):
        return value.default
    return value


def _extract_default(value: Any, make_optional: bool) -> Any:
    underlying_value = _extract_from_typer_parameter_info_if_necessary(value)
    if not make_optional:
        return underlying_value
    return underlying_value if underlying_value is not ... else None


def _is_optional(typ: type) -> TypeGuard[Type[Optional[Any]]]:
    return get_origin(typ) is Union and type(None) in get_args(typ)



from typing import Callable, Optional

from pyappconf import BaseConfig

from cliconf.ext_pyappconf import save_model
from tests.config import OVERRIDES_CONFIGS_DIR, PLAIN_CONFIGS_DIR
from tests.fixtures.app_settings import (
    SETTINGS_NESTED_CONFIG_YAML,
    SETTINGS_ONE_YAML,
    SETTINGS_TWO_PY,
)
from tests.fixtures.cliconfs import (
    NestedConfig,
    default_func_for_single_command_py,
    my_cli_func_two_py,
)


class ConfigOne(BaseConfig):
    a: str
    b: int
    c: float = 3.2

    _settings = SETTINGS_ONE_YAML


class ConfigOneOptional(ConfigOne):
    a: Optional[str] = None
    b: Optional[int] = None


def generate_config_one():
    settings = ConfigOne._settings.copy(custom_config_folder=PLAIN_CONFIGS_DIR)
    ConfigOneOptional(settings=settings).save()


def generate_config_one_with_overrides():
    ConfigOne(a="a from config", b=1000, c=45.6).save()


def custom_d_func(c: float) -> str:
    return f"custom {c}"


def generate_config_two_py():
    ConfigTwo = my_cli_func_two_py.model_cls
    settings = ConfigTwo._settings.copy(custom_config_folder=PLAIN_CONFIGS_DIR)
    obj = ConfigTwo(settings=settings)
    save_model(obj, my_cli_func_two_py)


def generate_config_two_py_with_overrides():
    ConfigTwo = my_cli_func_two_py.model_cls
    current_imports = ConfigTwo._settings.py_config_imports
    new_imports = [*current_imports, "from tests.generate import custom_d_func"]
    settings = ConfigTwo._settings.copy(
        py_config_imports=new_imports, custom_config_folder=OVERRIDES_CONFIGS_DIR
    )
    obj = ConfigTwo(c=123.4, d=custom_d_func, settings=settings)
    save_model(obj, my_cli_func_two_py)


class ConfigWithNestingOptional(BaseConfig):
    a: Optional[str] = None
    b: Optional[NestedConfig] = None

    _settings = SETTINGS_NESTED_CONFIG_YAML


def generate_config_with_nesting():
    settings = ConfigWithNestingOptional._settings.copy(
        custom_config_folder=PLAIN_CONFIGS_DIR
    )
    ConfigWithNestingOptional(settings=settings).save()


if __name__ == "__main__":
    generate_config_one()
    generate_config_one_with_overrides()
    generate_config_two_py()
    generate_config_two_py_with_overrides()
    generate_config_with_nesting()



import inspect
from typing import Any, Callable, List

import click
import typer.models
from typer.main import get_click_param, lenient_issubclass
from typer.utils import get_params_from_function
from typing_extensions import TypeGuard


def get_arg_names_that_can_be_processed_by_typer(
    callback: Callable[..., Any]
) -> List[str]:
    can_process: List[str] = []
    parameters = get_params_from_function(callback)
    for param_name, param in parameters.items():
        # Never process Click context
        if lenient_issubclass(param.annotation, click.Context):
            continue
        # Don't process arguments without types
        if param.annotation is inspect._empty:
            continue
        try:
            _, _ = get_click_param(param)
        except RuntimeError as e:
            if "Type not yet supported" in str(e):
                continue
            raise e
        else:
            can_process.append(param_name)
    return can_process


def is_typer_parameter_info(argument: Any) -> TypeGuard[typer.models.ParameterInfo]:
    return all(
        [
            hasattr(argument, "default"),
            hasattr(argument, "help"),
            hasattr(argument, "metavar"),
            hasattr(argument, "is_eager"),
        ]
    )



"""
Framework that lets you write CLIs with Typer that can also be configured via py-app-conf
"""
from cliconf.configure import configure, model_as_dict
from cliconf.ext_pyappconf import CLIAppConfig
from cliconf.main import CLIConf
from cliconf.settings import CLIConfSettings



from pathlib import Path

from pyappconf import ConfigFormats

from cliconf import CLIAppConfig
from tests.config import OVERRIDES_CONFIGS_DIR

SETTINGS_ONE_YAML = CLIAppConfig(
    app_name="MyApp",
    config_name="one",
    custom_config_folder=OVERRIDES_CONFIGS_DIR,
    default_format=ConfigFormats.YAML,
)

SETTINGS_ONE_RECURSIVE_YAML = CLIAppConfig(
    app_name="MyApp",
    config_name="one",
    # Ensure it should not find a default config
    custom_config_folder=Path("/non/existent/path"),
    default_format=ConfigFormats.YAML,
)


SETTINGS_TWO_PY = CLIAppConfig(
    app_name="MyApp",
    config_name="two",
    custom_config_folder=OVERRIDES_CONFIGS_DIR,
    default_format=ConfigFormats.PY,
    py_config_imports=[
        "from tests.fixtures.cliconfs import my_cli_func_two_py, default_func_for_single_command_py",
    ],
)

SETTINGS_ONE_MULTI_FORMAT = CLIAppConfig(
    app_name="MyApp",
    config_name="one",
    custom_config_folder=OVERRIDES_CONFIGS_DIR,
    default_format=ConfigFormats.YAML,
    multi_format=True,
)


SETTINGS_ALL_OPTIONAL_JSON = CLIAppConfig(
    app_name="MyApp",
    config_name="all_optional_json",
    custom_config_folder=OVERRIDES_CONFIGS_DIR,
    default_format=ConfigFormats.JSON,
)

SETTINGS_NESTED_CONFIG_YAML = CLIAppConfig(
    app_name="MyApp",
    config_name="nested_config",
    custom_config_folder=OVERRIDES_CONFIGS_DIR,
    default_format=ConfigFormats.YAML,
)



from pathlib import Path
from types import FunctionType
from typing import Optional, Type

from pyappconf import AppConfig, BaseConfig, ConfigFormats
from pydantic import Extra

from cliconf.dynamic_config_name import dynamic_model_class_name


def create_cli_base_config_class(
    base_cls: Type[BaseConfig], settings: Optional[AppConfig] = None
) -> Type[BaseConfig]:
    settings = settings or base_cls._settings
    prefix = _create_default_env_prefix(settings)

    class CLIBaseConfig(base_cls):  # type: ignore
        class Config:
            env_prefix = prefix
            extra = Extra.ignore

    return CLIBaseConfig


class CLIAppConfig(AppConfig):
    """
    Overrides some of the default settings for pyappconf to be more reasonable
    for what users would typically want to use in a CLI application.

    - Updates default folder to be the current directory
    - Outputs stub file for Python config format
    """

    def __init__(self, **kwargs):
        if "custom_config_folder" not in kwargs:
            kwargs["custom_config_folder"] = Path(".")
        if "py_config_generate_model_class_in_stub" not in kwargs:
            kwargs["py_config_generate_model_class_in_stub"] = True
        super().__init__(**kwargs)


def save_model(model: BaseConfig, func: FunctionType) -> None:
    if model.settings.default_format != ConfigFormats.PY:
        # Custom handling is only needed for Python config format
        # Other formats, just let pyappconf handle it
        model.save()
        return

    # Python config format
    # For .py file, add the dynamic config import to get the model class from the
    # command function.
    # For .pyi file, output as-is as it will have the dynamic config class defined statically
    if model.settings.py_config_create_stub:
        file_path = model.settings.config_location.with_suffix(".pyi")
        model.to_py_config_stub(
            file_path, py_config_stub_kwargs=dict(generate_model_class=True)
        )
    settings_with_dynamic_import = _add_dynamic_config_import_to_settings(
        model.settings, func
    )
    temp_model = model.copy(update=dict(settings=settings_with_dynamic_import))
    temp_model.to_py_config(model.settings.config_location, include_stub_file=False)


def _create_default_env_prefix(settings: AppConfig) -> str:
    return settings.app_name.replace("-", "_").replace(" ", "_").upper() + "_"


def _add_dynamic_config_import_to_settings(
    pyappconf_settings: AppConfig, func: FunctionType
) -> AppConfig:
    model_class_name = dynamic_model_class_name(func)
    py_config_dynamic_config_import = f"{model_class_name} = {func.__name__}.model_cls"
    if pyappconf_settings.py_config_imports is None:
        new_py_config_imports = [py_config_dynamic_config_import]
    else:
        new_py_config_imports = [
            *pyappconf_settings.py_config_imports,
            py_config_dynamic_config_import,
        ]
    return pyappconf_settings.copy(py_config_imports=new_py_config_imports)
